/* MCCP.c - Tcl_MCCP library by Cath Allen, Chris Allen.
 * Copyright 2000 by the above.
 *
 * Rewrite of Tcl_MCCP:
 * Copyright (C) 2002 Mark Patton
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This extension provides the "MCCP::init chan" command which
 * stacks a mccp protocol handler on top of chan and "MCCP::version chan"
 * which returns the MCCP version chan is currently using.
 */

#include <tcl.h>
#include <errno.h>
#include <string.h>
#include "mccpDecompress.h"

static Tcl_ChannelType mccp_type;

typedef struct mccp_chan {
  mc_state *mcinfo;
  Tcl_Channel self;
  Tcl_Channel base;
  Tcl_ChannelType *base_type;
  ClientData base_data;
} mccp_chan_t;

static void appendResult (Tcl_Interp *interp, ...)
{
  Tcl_Obj *obj;
  va_list ap;

  va_start (ap, interp);

  obj = Tcl_GetObjResult (interp);
  Tcl_AppendStringsToObjVA (obj, ap);

  va_end(ap);
}

static void mccp_chan_free (mccp_chan_t *chan)
{
  mudcompress_delete (chan->mcinfo);
  Tcl_Free ((char *) chan);
}

static int MCCP_Close (ClientData instanceData, Tcl_Interp *interp)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  mccp_chan_free (chan);
  return 0;
}

static int MCCP_Input (ClientData instanceData, char *buf, int buflen,
			 int *errorCodePtr)
{
  int written;
  const char *resp;
  char *mcbuf;
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverInputProc *base_input;

  if (mudcompress_error (chan->mcinfo)) {
    *errorCodePtr = EIO;
    return -1;
  }

  mcbuf = Tcl_Alloc (buflen);
  base_input = Tcl_ChannelInputProc (chan->base_type);
  written = base_input (chan->base_data , mcbuf, buflen, errorCodePtr);

/*    printf ("written: %d, buflen: %d\n", written, buflen); */
  
  if (written < 1) {
    Tcl_Free (mcbuf);
    return written;
  }

  mudcompress_receive (chan->mcinfo, mcbuf, written);
  Tcl_Free (mcbuf);
  resp = mudcompress_response (chan->mcinfo);

  if (resp) {
    int err;
    Tcl_DriverOutputProc *base_output;
    int n, towrite;

    base_output = Tcl_ChannelOutputProc (chan->base_type);

    written = 0;
    towrite = strlen (resp);

    do {
      n = base_output (chan->base_data, resp + written, towrite - written,
		       &err);

      if (n == -1) {
	/* FIXME: Do something if there's an error? */
	break;
      } else {
	written += n;
      }
    } while (written < towrite);
  }

  written = 0;

  do {
    written += mudcompress_get (chan->mcinfo, buf + written, buflen - written);
  } while (written < buflen && mudcompress_pending (chan->mcinfo));

  /*  printf ("written: %d, buflen: %d\n", written, buflen); */

  if (written == 0) {
    *errorCodePtr = EAGAIN;
    return -1;
  }

  return written;
}

/* the rest of these channel funcs simply call the funcs of the lower chan */

static int MCCP_BlockMode (ClientData instanceData, int mode)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverBlockModeProc *func = Tcl_ChannelBlockModeProc (chan->base_type);
  return func (chan->base_data, mode);
}

static int MCCP_Output (ClientData instanceData, char *buf, int toWrite,
		  int *errorCodePtr)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverOutputProc *func = Tcl_ChannelOutputProc (chan->base_type);
  return func (chan->base_data, buf, toWrite, errorCodePtr);
}

static int MCCP_Seek (ClientData instanceData, long offset, int seekMode,
		int *errorCodePtr)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverSeekProc *func = Tcl_ChannelSeekProc (chan->base_type);
  return func (chan->base_data, offset, seekMode, errorCodePtr);
}

static int MCCP_GetOption (ClientData instanceData, Tcl_Interp *interp,
		       char *optionName, Tcl_DString *dsPtr)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverGetOptionProc *func = Tcl_ChannelGetOptionProc (chan->base_type);
  return func (chan->base_data, interp, optionName, dsPtr);
}

static int MCCP_SetOption (ClientData instanceData, Tcl_Interp *interp, 
		     char *optionName, char *optionValue)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverSetOptionProc *func = Tcl_ChannelSetOptionProc (chan->base_type);
  return func (chan->base_data, interp, optionName, optionValue);
}

static void MCCP_Watch (ClientData instanceData, int mask)
{
 mccp_chan_t *chan = (mccp_chan_t *) instanceData;
 Tcl_DriverWatchProc *func = Tcl_ChannelWatchProc (chan->base_type);
 return func (chan->base_data, mask);
}

static int MCCP_GetHandle (ClientData instanceData, int direction,
		      ClientData *handlePtr)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverGetHandleProc *func = Tcl_ChannelGetHandleProc (chan->base_type);
  return func (chan->base_data, direction, handlePtr);
}

static int MCCP_Flush (ClientData instanceData)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverFlushProc *func = Tcl_ChannelFlushProc (chan->base_type);
  return func (chan->base_data);
}

static int MCCP_Handler (ClientData instanceData, int interestMask)
{
  mccp_chan_t *chan = (mccp_chan_t *) instanceData;
  Tcl_DriverHandlerProc *func = Tcl_ChannelHandlerProc (chan->base_type);
  return func (chan->base_data, interestMask);
}

static int MCCP_initCmd (ClientData clientData, Tcl_Interp *interp, int objc,
                Tcl_Obj *CONST objv[])
{
  Tcl_Channel base_chan;
  mccp_chan_t *chan;
  int mode;

  if (objc != 2) {
    Tcl_WrongNumArgs (interp, 1, objv, "sock");
    return TCL_ERROR;
  }

  base_chan = Tcl_GetChannel (interp, Tcl_GetString (objv[1]), &mode);

  if (base_chan == NULL) {
    appendResult (interp, "bad channel: ", Tcl_GetString (objv[1]),
		  (char *) 0);
    return TCL_ERROR;
  }

  /* register mccp channel type if not already registered */

  if (Tcl_ChannelInputProc (&mccp_type) == NULL) {
    mccp_type = *Tcl_GetChannelType (base_chan);

    mccp_type.closeProc = MCCP_Close;
    mccp_type.close2Proc = NULL;
    mccp_type.inputProc = MCCP_Input;
    mccp_type.outputProc = MCCP_Output;
    
    if (mccp_type.seekProc) {
      mccp_type.seekProc = MCCP_Seek;
    }

    if (mccp_type.setOptionProc) {
      mccp_type.setOptionProc = MCCP_SetOption;
    }

    if (mccp_type.getOptionProc) {
      mccp_type.getOptionProc = MCCP_GetOption;
    }
    
    if (mccp_type.handlerProc) {
      mccp_type.handlerProc = MCCP_Handler;
    }

    mccp_type.watchProc = MCCP_Watch;
    mccp_type.getHandleProc = MCCP_GetHandle;
    mccp_type.blockModeProc = MCCP_BlockMode;
    mccp_type.flushProc = MCCP_Flush;
  }

  chan = (mccp_chan_t *) Tcl_Alloc (sizeof (mccp_chan_t));
  chan->self = Tcl_StackChannel (interp, &mccp_type, (ClientData) chan,
				 mode, base_chan);
  if (chan->self == NULL) {
    Tcl_Free ((char *) chan);
    return TCL_ERROR;
  }
  
  chan->mcinfo = mudcompress_new ();
  chan->base = Tcl_GetStackedChannel (chan->self);
  chan->base_type = Tcl_GetChannelType (chan->base);
  chan->base_data = Tcl_GetChannelInstanceData (chan->base);

  return TCL_OK;
}

static int MCCP_versionCmd (ClientData clientData, Tcl_Interp *interp,
			    int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Channel real_chan;
  mccp_chan_t *chan;

  if (objc != 2) {
    Tcl_WrongNumArgs (interp, 1, objv, "chan");
    return TCL_ERROR;
  }

  real_chan = Tcl_GetChannel (interp, Tcl_GetString (objv[1]), 0);

  if (real_chan == NULL) {
    return TCL_ERROR;
  }

  real_chan = Tcl_GetTopChannel (real_chan);

  if (Tcl_GetChannelType (real_chan) != &mccp_type) {
    appendResult (interp, "not a mccp channel: ", Tcl_GetString (objv[1]),
		  (char *) 0);
    return TCL_ERROR;
  }

  chan = (mccp_chan_t *) Tcl_GetChannelInstanceData (real_chan);

  Tcl_SetObjResult (interp,
		    Tcl_NewIntObj (mudcompress_version (chan->mcinfo)));
  return TCL_OK;
}

int Mccp_Init (Tcl_Interp * interp) {
  if (!interp) {
    return TCL_ERROR;
  }

  if (!Tcl_InitStubs (interp, "8.3.0", 0)) {
    return TCL_ERROR;
  }

  if (Tcl_PkgProvide (interp, "MCCP", VERSION) != TCL_OK) {
    return TCL_ERROR;
  }

  Tcl_CreateObjCommand(interp, "MCCP::init", MCCP_initCmd,
		       (ClientData) NULL, (Tcl_CmdDeleteProc *)NULL);
  Tcl_CreateObjCommand(interp, "MCCP::version", MCCP_versionCmd,
		       (ClientData) NULL, (Tcl_CmdDeleteProc *)NULL);
 
  return TCL_OK;
}
